use anyhow::Result;
use serde_derive::Serialize;

use tracing::instrument;

use crate::{page::Page, tag::Tag, AppState};

#[derive(Serialize)]
struct FeedContext<'a> {
    feed_url: &'a str,
    base_url: &'a str,
    site_title: &'a str,
    last_updated: &'a str,
    tag: Option<&'a Tag>,
    posts: &'a [&'a Page],
}

#[instrument(skip(state))]
pub fn render_atom_feed(state: &AppState) -> Result<String> {
    let mut posts: Vec<_> = state
        .pages
        .values()
        .filter(|p| !p.draft && p.is_published())
        .collect();

    posts.sort_by_key(|p| &p.date);
    posts.reverse();
    posts.truncate(10);

    let updated = posts.iter().map(|p| p.updated.or(p.date)).max().flatten();
    let feed = FeedContext {
        feed_url: &format!("{}atom.xml", state.base_url),
        base_url: &state.base_url.to_string(),
        site_title: &state.settings.title,
        last_updated: &updated.map_or_else(String::default, |d| d.to_rfc3339()),
        tag: None,
        posts: &posts,
    };

    let ctx = tera::Context::from_serialize(feed)?;

    Ok(state.tera.render("atom.xml", &ctx)?)
}

#[instrument(skip(tag, state))]
pub fn render_atom_tag_feed(tag: &Tag, state: &AppState) -> Result<String> {
    let mut posts: Vec<_> = state
        .pages
        .values()
        .filter(|p| !p.draft && p.is_published() && p.tags.contains(&tag.slug))
        .collect();

    posts.sort_by_key(|p| &p.date);
    posts.reverse();
    posts.truncate(10);

    let updated = posts.iter().map(|p| p.updated.or(p.date)).max().flatten();
    let slug = &tag.slug;
    let feed = FeedContext {
        feed_url: &format!("{}tags/{}/atom.xml", state.base_url, slug),
        base_url: &state.base_url.to_string(),
        site_title: &state.settings.title,
        last_updated: &updated.map_or_else(String::default, |d| d.to_rfc3339()),
        tag: Some(tag),
        posts: &posts,
    };

    let ctx = tera::Context::from_serialize(feed)?;

    Ok(state.tera.render("atom.xml", &ctx)?)
}

struct _JsonFeed<'a> {
    version: &'a str,
    title: &'a str,
    home_page_url: &'a str,
    feed_url: &'a str,
    items: Vec<_JsonFeedItem<'a>>,
}

struct _JsonFeedItem<'a> {
    id: &'a str,
}
