use axum::{
    body::Body,
    extract::{Request, State},
    http::{header, HeaderMap, StatusCode},
    middleware::Next,
    response::{Html, IntoResponse, Response},
    routing::get,
    Router,
};
use chrono::{DateTime, FixedOffset};
use opentelemetry::{global, metrics::Counter, KeyValue};
use std::sync::Arc;
use tokio::sync::OnceCell;
use tower_http::services::ServeDir;
use tracing::{instrument, log::error};

use crate::{AppState, WebsiteError};

pub mod pages;
pub mod tags;

pub static HIT_COUNTER: OnceCell<Counter<u64>> = OnceCell::const_new();

async fn record_hit(method: String, path: String) {
    let counter = HIT_COUNTER
        .get_or_init(|| async { global::meter("tlxite").u64_counter("page_hit_count").init() })
        .await;

    counter.add(1, &[
        KeyValue::new("path", format!("{method} {path}")),
        ]);
}

pub fn routes() -> Router<Arc<AppState>> {
    Router::new()
        // .route("/", get(index))
        .merge(pages::router())
        .merge(tags::router())
        .route("/healthcheck", get(|| async { "OK" }))
        .nest_service("/static", ServeDir::new("static"))
        .fallback(|| async { WebsiteError::NotFound })
        .layer(axum::middleware::from_fn(metrics_middleware))
}

#[instrument(skip(state))]
pub async fn index(
    State(state): State<Arc<AppState>>,
    headers: HeaderMap,
) -> std::result::Result<Response, WebsiteError> {
    if let Some(res) = should_return_304(&headers, Some(state.startup_time.into())) {
        return Ok(res);
    }
    let mut ctx = tera::Context::new();
    ctx.insert("base_url", &state.base_url.to_string());
    let res = state.tera.render("index.html", &ctx).map_err(|e| {
        error!("Failed rendering index: {}", e);
        WebsiteError::InternalError(e.into())
    })?;
    Ok((
        StatusCode::OK,
        [
            (
                header::LAST_MODIFIED,
                state.startup_time.to_rfc2822().as_str(),
            ),
            (header::CACHE_CONTROL, "no-cache"),
        ],
        Html(res),
    )
        .into_response())
}

#[instrument(skip(request, next))]
pub async fn metrics_middleware(request: Request, next: Next) -> Response {
    let path = request.uri().path().to_string();
    let method = request.method().clone();

    let response = next.run(request).await;

    if !response.status().is_client_error() {
        record_hit(method.to_string(), path).await;
    }

    response
}

fn should_return_304(
    headers: &HeaderMap,
    last_changed: Option<DateTime<FixedOffset>>,
) -> Option<Response> {
    let Some(date) = last_changed else {
        return None;
    };
    let Some(since) = headers.get(header::IF_MODIFIED_SINCE) else {
        return None;
    };

    let Ok(parsed) = DateTime::<FixedOffset>::parse_from_rfc2822(since.to_str().unwrap()) else {
        return None;
    };

    if date >= parsed {
        Some(Response::builder().status(304).body(Body::empty()).unwrap())
    } else {
        None
    }
}

impl IntoResponse for WebsiteError {
    fn into_response(self) -> Response {
        match self {
            WebsiteError::NotFound => (StatusCode::NOT_FOUND, "not found").into_response(),
            WebsiteError::InternalError(e) => {
                if let Some(s) = e.source() {
                    error!("internal error: {}: {}", e, s);
                } else {
                    error!("internal error: {}", e);
                }
                (StatusCode::INTERNAL_SERVER_ERROR, "internal error").into_response()
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use std::sync::Arc;

    use crate::AppState;

    #[test]
    fn render_index() {
        let tera = tera::Tera::new("templates/**/*").unwrap();
        let mut ctx = tera::Context::new();
        ctx.insert("base_url", "http://localhost/");
        let _res = tera.render("index.html", &ctx).unwrap();
    }

    #[tokio::test]
    async fn setup_routes() {
        let mut state = AppState {
            startup_time: chrono::offset::Utc::now(),
            base_url: "http://localhost:8180".parse().unwrap(),
            tera: tera::Tera::new("templates/**/*").unwrap(),
            ..Default::default()
        };
        // Load the actual posts, just to make this test fail if
        // aliases overlap with themselves or other routes
        let posts = crate::page::load_all(&state, "pages/".into())
            .await
            .unwrap();
        state.tags = crate::tag::get_tags(posts.values());
        state.pages = posts;
        let state = Arc::new(state);

        super::routes().with_state(state).into_make_service();
    }
}
