use std::sync::Arc;

use anyhow::anyhow;
use axum::{
    body::Body,
    extract::{OriginalUri, Request, State},
    http::{self, header, HeaderMap, StatusCode, Uri},
    response::{Html, IntoResponse, Redirect, Response},
    routing::get,
    Router,
};

use serde_derive::Serialize;
use tower::ServiceExt;
use tower_http::services::ServeDir;
use tracing::instrument;

use crate::{
    page::{render_page, Page},
    AppState, WebsiteError,
};

use super::should_return_304;

pub fn router() -> Router<Arc<AppState>> {
    Router::new()
        .route("/atom.xml", get(feed))
        .route("/posts/", get(index))
        .route("/posts", get(|| async { Redirect::permanent("/posts/") }))
        .route("/", get(view))
        .route("/*path", get(view))
}

#[derive(Serialize, Debug)]
struct PageContext<'a> {
    title: &'a str,
}

#[instrument(skip(state, headers))]
async fn index(
    State(state): State<Arc<AppState>>,
    headers: HeaderMap,
) -> Result<Response, WebsiteError> {
    let mut posts: Vec<&Page> = state
        .pages
        .values()
        .filter(|p| !p.draft && p.is_published())
        .collect();

    let last_changed = posts.iter().filter_map(|p| p.last_modified()).max();

    if let Some(res) = should_return_304(&headers, last_changed) {
        return Ok(res);
    }

    posts.sort_by_key(|p| &p.date);
    posts.reverse();

    let ctx = PageContext { title: "Posts" };

    let mut c = tera::Context::new();
    c.insert("page", &ctx);
    c.insert("posts", &posts);
    c.insert("site_title", &state.settings.title);
    c.insert("base_url", &state.base_url.to_string());

    let res = state.tera.render("section_index.html", &c)?;

    Ok((
        StatusCode::OK,
        [
            (
                header::LAST_MODIFIED,
                last_changed
                    .map_or_else(|| state.startup_time.to_rfc2822(), |d| d.to_rfc2822())
                    .as_str(),
            ),
            (header::CACHE_CONTROL, "no-cache"),
        ],
        Html(res),
    )
        .into_response())
}

#[instrument(skip(state, uri, method, headers))]
async fn view(
    OriginalUri(uri): OriginalUri,
    State(state): State<Arc<AppState>>,
    method: http::method::Method,
    headers: HeaderMap,
) -> Result<Response, WebsiteError> {
    // Fetch post
    let Some(post) = state.pages.get(uri.path()) else {
        // Invalid path for a post, check aliases
        if let Some(p) = state.aliases.get(uri.path()) {
            return Ok(Redirect::permanent(p).into_response());
        }

        // No alias, check if we can do a simple redirect
        if !uri.path().ends_with('/') {
            let p = format!("{}/", uri.path());
            if state.pages.contains_key(&p) {
                return Ok(Redirect::permanent(&p).into_response());
            }
        }

        // no page, check if there's a static file to serve from here
        return get_static_file("pages", uri).await;
    };

    if !post.is_published() {
        return Err(WebsiteError::NotFound);
    }

    // Check if they have the current page cached
    if let Some(etag) = headers.get(header::IF_NONE_MATCH) {
        if etag.to_str().ok() == Some(post.etag.as_str()) {
            return Ok(StatusCode::NOT_MODIFIED.into_response());
        }
    }

    if method == http::method::Method::HEAD {
        return Ok((
            StatusCode::OK,
            [
                (header::ETAG, post.etag.as_str()),
                (header::CACHE_CONTROL, "no-cache"),
            ],
        )
            .into_response());
    }

    let res = render_page(&state, post).await?;

    Ok((
        StatusCode::OK,
        [
            (header::ETAG, post.etag.as_str()),
            (header::CACHE_CONTROL, "no-cache"),
        ],
        Html(res),
    )
        .into_response())
}

#[instrument(skip(state))]
pub async fn feed(
    State(state): State<Arc<AppState>>,
    headers: HeaderMap,
) -> Result<Response, WebsiteError> {
    let mut posts: Vec<&Page> = state
        .pages
        .values()
        .filter(|p| !p.draft && p.is_published())
        .collect();

    let last_changed = posts.iter().filter_map(|p| p.last_modified()).max();

    if let Some(res) = should_return_304(&headers, last_changed) {
        return Ok(res);
    }

    posts.sort_by_key(|p| &p.date);
    posts.reverse();
    posts.truncate(10);

    let last_modified =
        last_changed.map_or_else(|| state.startup_time.to_rfc2822(), |d| d.to_rfc2822());

    Ok((
        StatusCode::OK,
        [
            (header::CONTENT_TYPE, "application/atom+xml"),
            (header::LAST_MODIFIED, &last_modified),
        ],
        crate::feed::render_atom_feed(&state)?,
    )
        .into_response())
}

#[instrument]
async fn get_static_file(base_dir: &str, uri: Uri) -> Result<Response, WebsiteError> {
    let req = Request::builder().uri(uri).body(Body::empty()).unwrap();

    ServeDir::new(base_dir)
        .oneshot(req)
        .await
        .map(IntoResponse::into_response)
        .map_err(|e| anyhow!(e).into())
}

#[cfg(test)]
mod tests {
    use chrono::DateTime;

    use crate::page::Page;

    use super::PageContext;

    #[test]
    fn render_index() {
        let posts = vec![
            Page {
                title: Some("test".into()),
                slug: "test".into(),
                tags: vec!["abc".into(), "def".into()],
                date: Some(DateTime::parse_from_rfc3339("2023-03-26T13:04:01+02:00").unwrap()),
                ..Default::default()
            },
            Page {
                title: Some("test2".into()),
                slug: "test2".into(),
                date: None,
                ..Default::default()
            },
        ];
        let page = PageContext { title: "Posts" };
        let mut ctx = tera::Context::new();
        ctx.insert("base_url", "http://localhost/");
        ctx.insert("page", &page);
        ctx.insert("posts", &posts);

        let tera = tera::Tera::new("templates/**/*").unwrap();
        let _res = tera.render("posts_index.html", &ctx).unwrap();
    }
}
