#![warn(clippy::pedantic)]
use std::{collections::HashMap, fmt::Display, sync::Arc};

use axum::http::Uri;
use chrono::DateTime;

use page::Page;

use settings::Settings;
use tag::Tag;
use tera::Tera;

use tokio::{net::TcpListener, signal};
use tower_http::{compression::CompressionLayer, cors::CorsLayer};
use tracing::{debug, instrument, log::info};

use anyhow::{Error, Result};

mod feed;
mod handlers;
mod helpers;
mod hilighting;
mod markdown;
mod observability;
mod page;
mod settings;
mod tag;

#[derive(Default)]
pub struct AppState {
    startup_time: DateTime<chrono::offset::Utc>,
    base_url: Uri,
    pages: HashMap<String, Page>,
    aliases: HashMap<String, String>,
    settings: Settings,
    tags: HashMap<String, Tag>,
    tera: Tera,
}

#[tokio::main]
async fn main() -> Result<()> {
    let cfg = settings::get()?;
    observability::init(&cfg)?;

    info!("Starting server...");
    let listener = TcpListener::bind(&cfg.bind_address).await.unwrap();
    info!(
        "Bind address: {:?} Base Url: {:?}",
        cfg.bind_address, cfg.base_url
    );
    let app = init_app(cfg).await?;
    axum::serve(listener, app.into_make_service())
        .with_graceful_shutdown(shutdown_signal())
        .await?;

    opentelemetry::global::shutdown_tracer_provider();

    Ok(())
}

#[instrument(skip(cfg))]
async fn init_app(cfg: Settings) -> Result<axum::routing::Router> {
    let base_url: Uri = cfg.base_url.parse().unwrap();

    let tera = Tera::new("templates/**/*")?;
    let mut state = AppState {
        startup_time: chrono::offset::Utc::now(),
        base_url,
        tera,
        settings: cfg,
        ..Default::default()
    };

    let pages = page::load_all(&state, "pages/".into()).await?;
    info!("{} pages loaded", pages.len());
    for page in pages.values() {
        debug!(
            "slug: {:?}, path: {:?}, children: {:?}",
            page.slug, page.absolute_path, page.child_pages
        );
    }

    let tags = tag::get_tags(pages.values());
    state.aliases = pages
        .values()
        .flat_map(|p| {
            p.aliases
                .iter()
                .map(|a| (a.clone(), p.absolute_path.clone()))
        })
        .collect();

    state.pages = pages;
    state.tags = tags;

    Ok(handlers::routes()
        .layer(CorsLayer::permissive())
        .layer(CompressionLayer::new())
        .layer(
            tower_http::trace::TraceLayer::new_for_http()
                .make_span_with(observability::make_span)
                .on_response(observability::on_response),
        )
        .with_state(Arc::new(state)))
}

async fn shutdown_signal() {
    let ctrl_c = async {
        signal::ctrl_c()
            .await
            .expect("failed to install Ctrl+C handler");
    };

    #[cfg(unix)]
    let terminate = async {
        signal::unix::signal(signal::unix::SignalKind::terminate())
            .expect("failed to install signal handler")
            .recv()
            .await;
    };

    #[cfg(not(unix))]
    let terminate = std::future::pending::<()>();

    tokio::select! {
        () = ctrl_c => {},
        () = terminate => {},
    }
}

#[derive(Debug)]
pub enum WebsiteError {
    NotFound,
    InternalError(Error),
}

impl std::error::Error for WebsiteError {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            WebsiteError::NotFound => None,
            WebsiteError::InternalError(e) => Some(e.as_ref()),
        }
    }
}

impl Display for WebsiteError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            WebsiteError::NotFound => write!(f, "Not found"),
            WebsiteError::InternalError(e) => write!(f, "Internal error: {e}"),
        }
    }
}

impl From<tera::Error> for WebsiteError {
    fn from(value: tera::Error) -> Self {
        WebsiteError::InternalError(value.into())
    }
}

impl From<Error> for WebsiteError {
    fn from(value: Error) -> Self {
        WebsiteError::InternalError(value)
    }
}
