use crate::helpers;
use crate::hilighting;
use axum::http::Uri;
use cached::once_cell::sync::Lazy;
use pulldown_cmark::CodeBlockKind;
use pulldown_cmark::Event;
use pulldown_cmark::LinkType;
use pulldown_cmark::Tag;
use pulldown_cmark::TagEnd;
use pulldown_cmark::{Options, Parser};
use regex::Regex;
use tracing::instrument;

static STARTS_WITH_SCHEMA_RE: Lazy<Regex> = Lazy::new(|| Regex::new(r"^\w+:").unwrap());
static EMAIL_RE: Lazy<Regex> = Lazy::new(|| Regex::new(r"^.+?@\w+(\.\w+)*$").unwrap());

pub struct RenderResult {
    pub content_html: String,
    pub metadata: String,
}

#[instrument(skip(markdown))]
pub fn render_markdown_to_html(base_uri: Option<&Uri>, markdown: &str) -> RenderResult {
    let mut opt = Options::empty();
    opt.insert(Options::ENABLE_FOOTNOTES);
    opt.insert(Options::ENABLE_HEADING_ATTRIBUTES);
    opt.insert(Options::ENABLE_STRIKETHROUGH);
    opt.insert(Options::ENABLE_TABLES);
    opt.insert(Options::ENABLE_TASKLISTS);
    opt.insert(Options::ENABLE_SMART_PUNCTUATION);
    opt.insert(Options::ENABLE_PLUSES_DELIMITED_METADATA_BLOCKS);

    let mut content_html = String::with_capacity(markdown.len());
    let parser = Parser::new_ext(markdown, opt);

    let mut accumulated_block = String::new();
    let mut code_lang = None;
    let mut meta_kind = None;
    let mut events = Vec::new();
    let mut metadata = String::new();
    for event in parser {
        match event {
            Event::Text(text) => {
                if code_lang.is_some() || meta_kind.is_some() {
                    accumulated_block.push_str(&text);
                } else {
                    events.push(Event::Text(text));
                }
            }
            Event::Start(Tag::MetadataBlock(kind)) => {
                meta_kind = Some(kind);
            }
            Event::End(TagEnd::MetadataBlock(_)) => {
                meta_kind = None;
                metadata.push_str(&accumulated_block);
                accumulated_block.clear();
            }
            Event::Start(Tag::Link {
                mut dest_url,
                link_type,
                title,
                id,
            }) => {
                if let Some(uri) = base_uri {
                    if link_type != LinkType::Email
                        && !STARTS_WITH_SCHEMA_RE.is_match(&dest_url)
                        && !EMAIL_RE.is_match(&dest_url)
                    {
                        // convert relative URIs to absolute URIs
                        dest_url = helpers::uri_with_path(uri, &dest_url).to_string().into();
                    }
                }
                events.push(Event::Start(Tag::Link {
                    dest_url,
                    link_type,
                    title,
                    id,
                }));
            }
            Event::Start(Tag::Image {
                link_type,
                mut dest_url,
                title,
                id,
            }) => {
                if let Some(uri) = base_uri {
                    if !dest_url.contains("://") && !dest_url.contains('@') {
                        // convert relative URIs to absolute URIs
                        dest_url = helpers::uri_with_path(uri, &dest_url).to_string().into();
                    }
                }
                events.push(Event::Start(Tag::Image {
                    link_type,
                    dest_url,
                    title,
                    id,
                }));
            }
            Event::Start(Tag::CodeBlock(kind)) => {
                if let CodeBlockKind::Fenced(lang) = kind {
                    code_lang = Some(lang);
                }
            }
            Event::End(TagEnd::CodeBlock) => {
                let lang = code_lang.take().unwrap_or("".into());
                let res = hilighting::hilight(&accumulated_block, &lang, Some("base16-ocean.dark"))
                    .unwrap();

                events.push(Event::Html(res.into()));
                accumulated_block.clear();
            }
            _ => events.push(event),
        }
    }

    events.retain(|e| match e {
        Event::Text(t) | Event::Html(t) => !t.is_empty(),
        _ => true,
    });

    pulldown_cmark::html::push_html(&mut content_html, events.into_iter());

    RenderResult {
        content_html,
        metadata,
    }
}
