use std::{borrow::Cow, net::SocketAddr, time::Duration};

use anyhow::{Error, Result};
use axum::{
    extract::{ConnectInfo, MatchedPath, OriginalUri, Request},
    http::{header, HeaderMap},
    response::Response,
};
use opentelemetry::{global, KeyValue};
use opentelemetry_otlp::WithExportConfig;
use opentelemetry_sdk::{
    metrics::reader::{DefaultAggregationSelector, DefaultTemporalitySelector}, propagation::TraceContextPropagator, trace::{RandomIdGenerator, Sampler}, Resource
};
use tracing::{field::Empty, info_span, Span};
use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt, EnvFilter};

use crate::settings::Settings;

pub fn init(cfg: &Settings) -> Result<(), Error> {
    let filter = EnvFilter::builder()
        .with_env_var("TLX_LOG")
        .with_default_directive("info".parse()?)
        .parse_lossy(&cfg.logging);

    if cfg.otlp.enabled {
        let tracer = opentelemetry_otlp::new_pipeline()
            .tracing()
            .with_trace_config(
                opentelemetry_sdk::trace::config()
                    .with_sampler(Sampler::AlwaysOn)
                    .with_id_generator(RandomIdGenerator::default())
                    .with_resource(Resource::new(vec![KeyValue::new("service.name", "tlxite")])),
            )
            .with_exporter(
                opentelemetry_otlp::new_exporter()
                    .tonic()
                    .with_endpoint(&cfg.otlp.endpoint),
            )
            .install_batch(opentelemetry_sdk::runtime::Tokio)?;

        global::set_text_map_propagator(TraceContextPropagator::new());
        let otel_tracer = tracing_opentelemetry::layer().with_tracer(tracer);

        let meter = opentelemetry_otlp::new_pipeline()
            .metrics(opentelemetry_sdk::runtime::Tokio)
            .with_exporter(
                opentelemetry_otlp::new_exporter()
                    .tonic()
                    .with_endpoint(&cfg.otlp.endpoint),
            )
            .with_resource(Resource::new(vec![KeyValue::new("service.name", "tlxite")]))
            .with_period(Duration::from_secs(3))
            .with_timeout(Duration::from_secs(10))
            .with_aggregation_selector(DefaultAggregationSelector::new())
            .with_temporality_selector(DefaultTemporalitySelector::new())
            .build()?;

        global::set_meter_provider(meter);

        // let logger = opentelemetry_otlp::new_pipeline()
        //     .logging()
        //     .with_exporter(
        //         opentelemetry_otlp::new_exporter()
        //             .tonic()
        //             .with_endpoint(&cfg.otlp.endpoint),

        //     )
        //     .install_batch(opentelemetry_sdk::runtime::Tokio)?;

        tracing_subscriber::registry()
            .with(filter)
            .with(otel_tracer)
            .with(tracing_subscriber::fmt::layer().compact())
            .init();
    }
    else {
        tracing_subscriber::registry()
            .with(filter)
            .with(tracing_subscriber::fmt::layer().compact())
            .init();
    }

    Ok(())
}

pub fn make_span(req: &Request) -> Span {
    let uri = if let Some(OriginalUri(uri)) = req.extensions().get::<OriginalUri>() {
        uri
    } else {
        req.uri()
    };
    let path = uri.path();
    let route = req
        .extensions()
        .get::<MatchedPath>()
        .map_or(path, axum::extract::MatchedPath::as_str);
    let method = req.method().as_str();
    let scheme = req.uri().scheme().map_or("HTTP", |s| s.as_str());

    let user_agent = req
        .headers()
        .get(header::USER_AGENT)
        .map_or("", |h| h.to_str().unwrap_or(""));


    let client_ip = parse_x_forwarded_for(req.headers())
        .or_else(|| {
            req.extensions()
                .get::<ConnectInfo<SocketAddr>>()
                .map(|ConnectInfo(client_ip)| Cow::from(client_ip.to_string()))
        })
        .unwrap_or_default();

    info_span!(
        "http request",
        otel.name = %format!("{method} {route}"),
        otel.kind = &"server",
        client.address = %client_ip,
        http.route = %route,
        http.request.method = %method,
        url.scheme = %scheme,
        url.path = %path,
        url.query = %uri.query().unwrap_or_default(),
        user_agent.original = %user_agent,
        http.response.status_code = Empty,
        otel.status_code = Empty,
    )
}

fn parse_x_forwarded_for(headers: &HeaderMap) -> Option<Cow<'_, str>> {
    let value = headers.get("x-forwarded-for")?;
    let value = value.to_str().ok()?;
    let mut ips = value.split(',');
    Some(ips.next()?.trim().into())
}

pub fn on_response(response: &Response, _latency: Duration, span: &Span) {
    span.record("http.response.status_code", response.status().as_str());
    if response.status().is_server_error() {
        span.record(
            "otel.status_code",
            if response.status().is_server_error() {
                "ERROR"
            } else {
                "OK"
            },
        );
    }
}
