use std::{
    collections::HashMap,
    ffi::OsStr,
    fmt::Debug,
    hash::{Hash, Hasher},
    path::{Path, PathBuf},
};

use anyhow::Result;

use chrono::{DateTime, FixedOffset};

use serde_derive::{Deserialize, Serialize};
use tokio::fs;

use tracing::{instrument, log::debug};

use crate::{helpers, markdown, AppState, WebsiteError};

#[derive(Deserialize, Debug, Default)]
pub struct TomlFrontMatter {
    pub title: Option<String>,
    pub date: Option<toml::value::Datetime>,
    pub updated: Option<toml::value::Datetime>,
    pub draft: Option<bool>,
    pub template: Option<String>,
    pub aliases: Option<Vec<String>>,
    pub tags: Option<Vec<String>>,
}

#[derive(Serialize, Clone, Debug, Default)]
pub struct Page {
    pub title: Option<String>,
    pub draft: bool,
    pub date: Option<DateTime<FixedOffset>>,
    pub updated: Option<DateTime<FixedOffset>>,
    pub aliases: Vec<String>,
    pub tags: Vec<String>,
    pub child_pages: Vec<String>,
    pub content: String,
    pub template: String,
    pub slug: String,
    pub absolute_path: String,
    pub section: Option<String>,
    pub etag: String,
}

impl Page {
    pub fn new(slug: String, content: String, fm: TomlFrontMatter) -> Page {
        let mut hasher = std::hash::DefaultHasher::default();
        fm.title.hash(&mut hasher);
        fm.draft.hash(&mut hasher);
        fm.tags.hash(&mut hasher);
        content.hash(&mut hasher);
        let etag = format!("W/\"{:x}\"", hasher.finish());

        Page {
            absolute_path: if slug.is_empty() {
                String::from("/")
            } else {
                format!("/{slug}/")
            },
            slug,
            etag,
            section: None,
            content,
            child_pages: vec![],
            template: fm.template.unwrap_or_else(|| "page.html".to_string()),
            title: fm.title,
            draft: fm.draft.unwrap_or(false),
            date: fm
                .date
                .map(|d| DateTime::parse_from_rfc3339(&d.to_string()).expect("bad toml datetime")),
            updated: fm
                .updated
                .map(|d| DateTime::parse_from_rfc3339(&d.to_string()).expect("bad toml datetime")),
            aliases: fm.aliases.unwrap_or_default(),
            tags: fm.tags.unwrap_or_default(),
        }
    }

    pub fn is_published(&self) -> bool {
        let now = chrono::offset::Local::now();
        if let Some(date) = self.date {
            return date.timestamp() - now.timestamp() <= 0;
        }
        true
    }

    pub fn last_modified(&self) -> Option<DateTime<FixedOffset>> {
        self.updated.or(self.date)
    }
}

#[instrument(skip(state))]
pub async fn load_all(state: &AppState, folder: PathBuf) -> Result<HashMap<String, Page>> {
    let mut pages = HashMap::<String, Page>::new();
    let mut dirs: Vec<PathBuf> = vec![folder.clone()];
    let mut section_stack: Vec<String> = vec![];

    while let Some(dir) = dirs.pop() {
        let mut read_dir = fs::read_dir(&dir).await?;

        let mut section_index = dir.clone();
        section_index.push("index.md");
        let is_section = section_index.exists();
        let current_section = if is_section {
            let mut page = load_page(state, &section_index, &folder).await?;
            page.section = section_stack.last().cloned();
            section_stack.push(page.absolute_path.clone());
            pages.insert(page.absolute_path.clone(), page);
            section_stack.last()
        } else {
            section_stack.last()
        };
        debug!("current section: {current_section:?}");

        let mut child_pages = vec![];

        while let Some(entry) = read_dir.next_entry().await? {
            let path = entry.path();
            if path.is_dir() {
                dirs.push(path);
            } else if let Some(ext) = path.extension() {
                if ext == "md" {
                    // it's a page to load
                    if path.file_name() == Some(OsStr::new("index.md")) {
                        // we've already loaded the section index page
                        continue;
                    }
                    let mut page = load_page(state, &path, &folder).await?;
                    page.section = current_section.cloned();
                    child_pages.push(page.absolute_path.clone());
                    pages.insert(page.absolute_path.clone(), page);
                }
            }
        }

        if let Some(section) = current_section {
            if let Some(section) = pages.get_mut(section) {
                section.child_pages.append(&mut child_pages);
            }
        }
    }

    Ok(pages)
}

#[instrument(skip(state))]
pub async fn load_page(state: &AppState, path: &Path, root_folder: &Path) -> Result<Page> {
    debug!("loading page: {path:?}");

    let content = fs::read_to_string(path).await?;

    let path_str = path.to_string_lossy().replace('\\', "/");
    let root = root_folder.to_string_lossy();

    let slug = path_str[root.len()..]
        .trim_start_matches('/')
        .trim_end_matches(".html")
        .trim_end_matches(".md")
        .trim_end_matches("index")
        .trim_end_matches('/');

    let base_path = if let Some(i) = path_str.rfind('/') {
        &path_str[root.len()..=i]
    } else {
        &path_str[root.len()..]
    };

    let base_uri = helpers::uri_with_path(&state.base_url, base_path);

    let content = markdown::render_markdown_to_html(Some(&base_uri), &content);

    Ok(Page::new(
        slug.to_string(),
        content.content_html,
        toml::from_str(&content.metadata)?,
    ))
}

#[instrument(skip(state, page))]
pub async fn render_page(state: &AppState, page: &Page) -> Result<String, WebsiteError> {
    let mut ctx = tera::Context::new();
    ctx.insert("page", &page);
    ctx.insert("all_pages", &state.pages);
    ctx.insert("site_title", &state.settings.title);
    ctx.insert("base_url", &state.base_url.to_string());

    state
        .tera
        .render(&page.template, &ctx)
        .map_err(std::convert::Into::into)
}

#[cfg(test)]
mod tests {
    use tera::Tera;

    use crate::AppState;

    #[tokio::test]
    async fn render_all_posts() {
        let mut state = AppState {
            base_url: "http://localhost:8180".parse().unwrap(),
            tera: Tera::new("templates/**/*").unwrap(),
            ..Default::default()
        };

        state.pages = super::load_all(&state, "pages/".into()).await.unwrap();
        for post in state.pages.values() {
            super::render_page(&state, post).await.unwrap();
        }
    }
}
